# Copyright 2018 John Reese
# Licensed under the MIT license

import asyncio
import logging
from typing import cast, Any, AsyncIterator

import aiohttp

from .state import Cache
from .types import Auto, Channel, Event, Group, User, Response, RTMStart

log = logging.getLogger(__name__)


class SlackError(Exception):
    """
    Generic error type for all Slack-related errors.
    """

    def __init__(self, message: str, *context: Any) -> None:
        super().__init__(message)
        self.context = context


class Slack:
    """
    Slack API entry point.
    """

    def __init__(self, token: str) -> None:
        self.token: str = token
        self.session = aiohttp.ClientSession(
            headers={"Authorization": f"Bearer {self.token}"}
        )

        self.me: Auto = Auto()
        self.team: Auto = Auto()
        self.channels = Cache(Channel, "channels.info")
        self.users = Cache(User, "users.info")
        self.groups = Cache(Group, "groups.info")

    def __del__(self) -> None:
        asyncio.ensure_future(self.close())

    async def __aenter__(self) -> "Slack":
        return self

    async def __aexit__(self, *args) -> None:
        await self.close()

    async def close(self) -> None:
        # TODO: track RTM sessions as tasks and cancel them here
        await self.session.close()

    async def api(self, method: str, **kwargs: str) -> Auto:
        async with self.session.post(
            f"https://slack.com/api/{method}", data=kwargs
        ) as request:
            if request.status != 200:
                raise SlackError(f"{method} returned status {request.status}")

            value = await request.json()
            if "self" in value:
                value["self_"] = value.pop("self")
            response = Response.generate(value, recursive=False)
            if not response.ok:
                raise SlackError(
                    f'{method} error: "{response.error}"', kwargs, response
                )
            if response.warning:
                log.warning(f'{method} warning: "{response.warning}"')
            return response

    async def rtm(self) -> AsyncIterator[Event]:
        """Connect to the realtime event API and start yielding events."""
        response = cast(RTMStart, await self.api("rtm.start"))

        self.me = Auto.generate(response.self_, "Me", recursive=False)
        self.team = Auto.generate(response.team, "Team", recursive=False)
        self.channels.fill(Channel.build(item) for item in response.channels)
        self.users.fill(User.build(item) for item in response.users)
        self.groups.fill(Group.build(item) for item in response.groups)

        log.debug(
            f"received {len(self.users)} users, {len(self.channels)} channels "
            f"and {len(self.groups)} groups from rtm.start"
        )

        async with self.session.ws_connect(response["url"]) as ws:
            async for msg in ws:
                event: Event = Event.generate(msg.json(), recursive=False)

                if event.type == "goodbye":
                    break

                yield event



# Copyright 2018 John Reese
# Licensed under the MIT license

"""
Basic type definitions for the Slack API.

Based on the Object Types documentation at https://api.slack.com/types
"""

import logging

from typing import Any, Dict, List, Mapping, Type, TypeVar
from attr import asdict, dataclass, fields_dict, ib, make_class

T = TypeVar("T", bound="Auto")

Generic = Mapping[str, Any]  # subvalues that we don't (yet?) care about
URL = str
HTML = str

log = logging.getLogger(__name__)


class Auto:

    def __init__(self, **kwargs) -> None:
        # silence mypy by have a default constructor
        pass

    def __contains__(self, key: str) -> bool:
        return key in asdict(self)

    def __getitem__(self, key: str) -> Any:
        return asdict(self)[key]

    @classmethod
    def build(cls: Type[T], data: Generic) -> T:
        """Build objects from dictionaries, recursively."""
        fields = fields_dict(cls)
        kwargs: Dict[str, Any] = {}
        for key, value in data.items():
            if key in fields:
                if isinstance(value, Mapping):
                    t = fields[key].type
                    if issubclass(t, Auto):
                        value = t.build(value)
                    else:
                        value = Auto.generate(value, name=key.title())
                kwargs[key] = value
            else:
                log.debug(f"got unknown attribute {key} for {cls.__name__}")
        return cls(**kwargs)

    @classmethod
    def generate(
        cls: Type[T], data: Generic, name: str = None, *, recursive: bool = True
    ) -> T:
        """Build dataclasses and objects from dictionaries, recursively."""
        if name is None:
            name = cls.__name__
        kls = make_class(name, {k: ib(default=None) for k in data}, bases=(cls,))
        data = {
            k: (
                cls.generate(v, k.title())
                if recursive and isinstance(v, Mapping)
                else v
            )
            for k, v in data.items()
        }
        return kls(**data)


@dataclass
class Value(Auto):
    value: str = ""
    creator: str = ""
    last_set: int = 0


@dataclass
class Channel(Auto):
    id: str
    name: str
    is_channel: bool = True
    created: int = 0
    creator: str = ""
    is_archived: bool = False
    is_general: bool = False
    name_normalized: str = ""
    is_shared: bool = False
    is_org_shared: bool = False
    is_member: bool = False
    is_private: bool = False
    is_mpim: bool = False
    last_read: str = ""
    latest: Generic = {}
    unread_count: int = 0
    unread_count_display: int = 0
    members: List[str] = []
    topic: Value = Value()
    purpose: Value = Value()
    previous_names: List[str] = []


@dataclass
class Conversation(Auto):
    id: str
    name: str
    is_channel: bool = False
    is_group: bool = False
    is_im: bool = False
    created: int = 0
    creator: str = ""
    is_archived: bool = False
    is_general: bool = False
    unlinked: int = 0
    name_normalized: str = ""
    is_read_only: bool = False
    is_shared: bool = False
    is_ext_shared: bool = False
    is_org_shared: bool = False
    pending_shared: List[Any] = []
    is_pending_ext_shared: bool = False
    is_member: bool = False
    is_private: bool = False
    is_mpim: bool = False
    last_read: str = ""
    topic: Value = Value()
    purpose: Value = Value()
    previous_names: List[str] = []
    num_members: int = 0
    locale: str = "en-US"


@dataclass
class Event(Auto):
    type: str = ""


@dataclass
class EventWrapper(Auto):
    token: str = ""
    team_id: str = ""
    api_app_id: str = ""
    event: Event = Event()
    type: str = ""
    authed_users: List[str] = []
    event_id: str = ""
    event_time: int = 0


@dataclass
class File(Auto):
    id: str
    created: int = 0
    timestamp: int = 0
    name: str = ""
    title: str = ""
    mimetype: str = ""
    filetype: str = ""
    pretty_type: str = ""
    user: str = ""
    mode: str = ""
    editable: bool = False
    is_external: bool = False
    external_type: str = ""
    username: str = ""
    size: int = 0
    url_private: URL = ""
    url_private_download: URL = ""
    thumb_64: URL = ""
    thumb_80: URL = ""
    thumb_360: URL = ""
    thumb_360_gif: URL = ""
    thumb_360_w: int = 0
    thumb_360_h: int = 0
    thumb_480: URL = ""
    thumb_480_w: int = 0
    thumb_480_h: int = 0
    thumb_160: URL = ""
    permalink: URL = ""
    permalink_public: URL = ""
    edit_link: URL = ""
    preview: HTML = ""
    preview_highlight: HTML = ""
    lines: int = 0
    lines_more: int = 0
    is_public: bool = False
    public_url_shared: bool = False
    display_as_bot: bool = False
    channels: List[str] = []
    groups: List[str] = []
    ims: List[str] = []
    initial_comment: Generic = {}
    num_stars: int = 0
    is_starred: bool = False
    pinned_to: List[str] = []
    reactions: List[Generic] = []
    comments_count: int = 0


@dataclass
class Group(Auto):
    id: str
    name: str
    is_group: bool = True
    created: int = 0
    creator: str = ""
    is_archived: bool = False
    is_mpim: bool = False
    members: List[str] = []
    topic: Value = Value()
    purpose: Value = Value()
    last_read: str = ""
    latest: Generic = {}
    unread_count: int = 0
    unread_count_display: int = 0


@dataclass
class IM(Auto):
    id: str
    is_im: bool = True
    user: str = ""
    created: int = 0
    is_user_deleted: bool = False


@dataclass
class MPIM(Auto):
    id: str
    name: str
    is_mpim: bool = False
    is_group: bool = False
    created: int = 0
    creator: str = ""
    members: List[str] = []
    last_read: str = ""
    latest: Generic = {}
    unread_count: int = 0
    unread_count_display: int = 0


@dataclass
class Profile(Auto):
    avatar_hash: str = ""
    status_text: str = ""
    status_emoji: str = ""
    real_name: str = ""
    display_name: str = ""
    real_name_normalized: str = ""
    display_name_normalized: str = ""
    email: str = ""
    image_24: URL = ""
    image_32: URL = ""
    image_48: URL = ""
    image_72: URL = ""
    image_192: URL = ""
    image_512: URL = ""
    team: str = ""


@dataclass
class Response(Auto):
    ok: bool
    error: str = ""
    warning: str = ""


@dataclass
class RTMStart(Auto):
    self_: Generic
    team: Generic
    channels: List
    users: List
    groups: List


@dataclass
class User(Auto):
    id: str
    team_id: str
    name: str
    deleted: bool = False
    color: str = ""
    real_name: str = ""
    tz: str = ""
    tz_label: str = ""
    tz_offset: int = 0
    profile: Profile = Profile()
    is_admin: bool = False
    is_owner: bool = False
    is_primary_owner: bool = False
    is_restricted: bool = False
    is_ultra_restricted: bool = False
    is_bot: bool = False
    is_stranger: bool = False
    updated: int = 0
    is_app_user: bool = False
    has_2fa: bool = False
    locale: str = "en-US"


@dataclass
class UserGroup(Auto):
    id: str
    team_id: str = ""
    is_usergroup: bool = True
    name: str = ""
    description: str = ""
    handle: str = ""
    is_external: bool = False
    date_create: int = 0
    date_update: int = 0
    date_delete: int = 0
    auto_type: str = ""
    created_by: str = ""
    updated_by: str = ""
    deleted_by: str = ""
    prefs: Generic = {}
    users: List[str] = []
    user_count: str = ""



# Copyright 2018 John Reese
# Licensed under the MIT license

"""
State management of the Slack connectionself.
"""

from typing import Dict, Iterator, Iterable, Mapping, Type, TypeVar, Optional

from .types import Auto

VT = TypeVar("VT", bound=Auto)


class Cache:
    """
    Cache objects of the given type, with optional readthrough via URL.
    """

    def __init__(self, _type: Type[VT], url: str = None) -> None:
        self.cache: Dict[str, VT] = {}
        self.by_name: Dict[str, str] = {}
        self.type = _type
        self.url = url

    def __iter__(self) -> Iterator[str]:
        return self.cache.__iter__()

    def __len__(self) -> int:
        return self.cache.__len__()

    def __contains__(self, key: str) -> bool:
        return key in self.cache or key in self.by_name

    def __getitem__(self, key: str) -> VT:
        if key in self.cache:
            return self.cache[key]

        if key in self.by_name:
            return self[self.by_name[key]]

        raise KeyError(f"{self.type.__name__} {key} not in cache")
        # TODO: make API request to fill cache?

    def __setitem__(self, key: str, value: VT) -> None:
        if not isinstance(value, self.type):
            raise ValueError(f"{key} is not {self.type.__name__}")
        self.cache[key] = value
        if "name" in value:
            self.by_name[value["name"]] = key

    def __delitem__(self, key: str) -> None:
        return self.cache.__delitem__(key)

    def keys(self) -> Iterable[str]:
        return self.cache.keys()

    def values(self) -> Iterable[VT]:
        return self.cache.values()

    def get(self, key: str, default: Optional[VT] = None) -> Optional[VT]:
        if key in self.cache:
            return self.cache[key]

        key = self.by_name.get(key, key)
        return self.cache.get(key, default)

    def fill(self, values: Iterable[VT], *, key: str = "id") -> None:
        for value in values:
            if not isinstance(value, self.type):
                raise ValueError(f"{value[key]} is not {self.type.__name__}")
            self.cache[value[key]] = value
            if "name" in value:
                self.by_name[value["name"]] = value[key]

    def update(self, values: Mapping[str, VT]) -> None:
        for key in values:
            value = values[key]
            if not isinstance(value, self.type):
                raise ValueError(f"{key} is not {self.type.__name__}")
            self.cache[key] = value
            if "name" in value:
                self.by_name[value["name"]] = value["id"]



# Copyright 2018 John Reese
# Licensed under the MIT license

"""AsyncIO Slack API"""

__version__ = "0.5.0"

from .core import Slack, SlackError
from .types import (
    Channel,
    Conversation,
    Event,
    EventWrapper,
    File,
    Group,
    IM,
    MPIM,
    Profile,
    RTMStart,
    User,
    UserGroup,
    Value,
)



from setuptools import setup

with open("README.md") as f:
    readme = f.read()

with open("requirements.txt") as f:
    requires = [line.strip() for line in f if line]

with open("aioslack/__init__.py") as f:
    for line in f:
        if line.startswith("__version__"):
            version = line.split('"')[1]

setup(
    name="aioslack",
    description="asyncio slack api",
    long_description=readme,
    long_description_content_type="text/markdown",
    version=version,
    author="John Reese",
    author_email="john@noswap.com",
    url="https://github.com/jreese/aioslack",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries",
    ],
    license="MIT",
    packages=["aioslack"],
    setup_requires=["setuptools>=38.6.0"],
    install_requires=requires,
)
